self.__BUILD_MANIFEST=function(s,c){return{__rewrites:{afterFiles:[],beforeFiles:[],fallback:[]},"/":[s,"static/chunks/541-eeb48182f95cfcc3.js",c,"static/css/b7bd5c40cc94be5a.css","static/chunks/pages/index-eb93a0e614138a7c.js"],"/_error":["static/chunks/pages/_error-7a92967bea80186d.js"],"/benchmarks":[s,c,"static/css/c9faa5ef13ea61d7.css","static/chunks/pages/benchmarks-76630194a3923726.js"],sortedPages:["/","/_app","/_error","/benchmarks"]}}("static/chunks/424-ca7471161a310844.js","static/chunks/811-348a36ae501321a4.js"),self.__BUILD_MANIFEST_CB&&self.__BUILD_MANIFEST_CB();